/**
 * Akeeba Remote Control
 * A utility to remotely control Akeeba Backup and Admin Tools
 * 
 * @version $Id$
 * @copyright 2008-2010 Nicholas K. Dionysopoulos / AkeebaBackup.com
 * @license GNU/GPL v3 or later
 */

function AkeebaBackupAPI()
{
	
}

AkeebaBackupAPI.prototype.errormessages = {
	200: "No error",
	401: "Wrong secret word",
	403: "Access denied",
	404: "Not found",
	405: "Invalid JSON API method",
	500: "Server reported an error",
	501: "Feature not implemented",
	503: "The remote feature is not activated in Akeeba Backup's Parameters or you have not specified a secret word",
	600: "Invalid JSON data"
};
	
// --- Site configuration
/** @var String The UUID of this site */
AkeebaBackupAPI.prototype.uuid = null;
/** @var String The URL to the site, e.g. http://www.example.com */
AkeebaBackupAPI.prototype.URL = '';
/** @var int The selected backup profile */
AkeebaBackupAPI.prototype.profile = 0;
/** @var String The secret work used for authenticating requests */
AkeebaBackupAPI.prototype.secret = '';
/** @var int The last backup's ID */
AkeebaBackupAPI.prototype.backupID = null;
/** @var Boolean Should I download the backup archives after backup? */
AkeebaBackupAPI.prototype.download = null;
/** @var String Where should I download everything? */
AkeebaBackupAPI.prototype.downloadDirectory = null;
/** @var int Should I delete the files after download? */
AkeebaBackupAPI.prototype.deleteAfterDownload = null;
/** @var Object The nex segment to download */
AkeebaBackupAPI.prototype.nextSegment = null;
/** @var int How many time I retried downloading this segment */
AkeebaBackupAPI.prototype.downloadRetries = 0;

// --- Data caches
AkeebaBackupAPI.prototype.cancelBackup = false; // When set to true, the backup will be cancelled
AkeebaBackupAPI.prototype.cancelDownload = false; // When set to true, the download will be cancelled
AkeebaBackupAPI.prototype.errorMessage = null;
AkeebaBackupAPI.prototype.versionCache = null;
AkeebaBackupAPI.prototype.profilesCache = null;
AkeebaBackupAPI.prototype.backupInfo = null;
AkeebaBackupAPI.prototype.log = ''; // Holds the log lines

/** @var int The last backup ID we were trying to download */
AkeebaBackupAPI.prototype.DownloadBackupID = null;
/** @var int The information of the backup record we were (are?) trying to download */
AkeebaBackupAPI.prototype.DownloadBackupInformation = null;
/** @var int The total size (in bytes) of all backup parts to be downloaded */
AkeebaBackupAPI.prototype.DownloadTotalSize = null;
/** @var int The size (in bytes) of all parts and fragments we have downloaded so far */
AkeebaBackupAPI.prototype.DownloadDownloadedSize = null;
/** @var Array An array of objects, defining the segment download queue */
AkeebaBackupAPI.prototype.DownloadQueue = null;
/** @var String The absolute path to the filename which will hold our download file */
AkeebaBackupAPI.prototype.DownloadOutputFilename = null;
/** @var Boolean Should I delete the files after download?  */
AkeebaBackupAPI.prototype.DownloadDeleteAfter = null;

/** @var Boolean When this flag is set, ARC will start a backup right after fetching the API version information */
AkeebaBackupAPI.prototype.startBackupAfterGetProfiles = false;

/** @var int The minimum API level supported by this application */
AkeebaBackupAPI.prototype.minimumAPIVersion = 318;

/**
 * Reset the internal log
 */
AkeebaBackupAPI.prototype.clearLog = function()
{
	this.log = '';
};

/**
 * Adds a log message
 * @param message String The log line to append to the log
 */
AkeebaBackupAPI.prototype.addLog = function(message)
{
	var mydate = date('Y-m-d H:i:s');
	this.log += mydate + ' | ' + message + "\n";
	Titanium.API.debug(message);
};

/**
 * Returns the raw log lines
 * @return String
 */
AkeebaBackupAPI.prototype.getLog = function()
{
	return this.log;
}

/**
 * Loads from a site definition
 * @param uuid String The site UUID to load
 * @return bool
 */
AkeebaBackupAPI.prototype.loadSite = function(uuid)
{	
	var siteDef = siteManager.getSiteDefinition(uuid);
	if(siteDef === false) return false;
	
	this.uuid = uuid;
	this.URL = siteDef.URL;
	this.profile = siteDef.profile;
	this.secret = siteDef.secret;
	
	this.download = siteDef.download;
	this.downloadDirectory = siteDef.downloadDirectory;
	
	this.deleteAfterDownload = siteDef.deleteAfterDownload;
	if(empty(this.deleteAfterDownload)) this.deleteAfterDownload = 0;
	
	return true;
};

/**
 * Performs an AJAX request using a prepared GET URL
 * @param url String The URL to load
 * @param successCallback function
 * @param errorCallback function
 * @return null
 */
AkeebaBackupAPI.prototype.doAJAX = function(url, successCallback, errorCallback, timeout)
{
	if(empty(timeout)) {
		timeout = 600000;
	}
	
	this.addLog('AJAX GET to ' + url);
	
	var uuid = this.uuid;
	var akeebaAPI = this;
	var structure =
	{
		type: "POST",
		url: url,
		cache: false,
		data: {},
		timeout: timeout,
		success: function(msg) {
			// Initialize
			var junk = null;
			var message = "";
			
			// Get rid of junk before the data
			var valid_pos = msg.indexOf('###');
			if( valid_pos == -1 ) {
				// Valid data not found in the response
				msg = 'Non-JSON data received: ' + msg;
				akeebaAPI.addLog(msg);
				if(errorCallback == null)
				{
					if(akeeba_error_callback != null)
					{
						akeeba_error_callback(msg, uuid);
					}
				}
				else
				{
					errorCallback(msg, uuid);
				}
				return;
			} else if( valid_pos != 0 ) {
				// Data is prefixed with junk
				junk = msg.substr(0, valid_pos);
				message = msg.substr(valid_pos);
			}
			else
			{
				message = msg;
			}
			message = message.substr(3); // Remove triple hash in the beginning
			
			// Get of rid of junk after the data
			var valid_pos = message.lastIndexOf('###');
			message = message.substr(0, valid_pos); // Remove triple hash in the end
			
			try {
				var data = JSON.parse(message);
			} catch(err) {
				var msg = err.message + "\n<br/>\n<pre>\n" + message + "\n</pre>";
				akeebaAPI.addLog('Invalid JSON data. Error data:');
				akeebaAPI.addLog(msg);
				if(errorCallback == null)
				{
					if(akeeba_error_callback != null)
					{
						akeeba_error_callback(msg, uuid);
					}
				}
				else
				{
					errorCallback(msg, uuid);
				}
				return;
			}
			
			var logData = '(too long; skipping dump)';
			if(strlen(message) < 65536) logData = message;
			akeebaAPI.addLog('AJAX GET successful. Data: ' + logData);
			// Call the callback function
			successCallback(data, uuid);
		},
		error: function(Request, textStatus, errorThrown) {
			var message = '<strong>Remote API Call Error</strong><br/>HTTP Status: '+Request.status+' ('+Request.statusText+')<br/>';
			message = message + 'Internal status: '+textStatus+'<br/>';
			message = message + 'XHR ReadyState: ' + Request.readyState + '<br/>';
			message = message + 'Raw server response:<br/>'+Request.responseText;

			akeebaAPI.addLog('AJAX GET failed. Error data:');
			akeebaAPI.addLog(message);

			if(errorCallback == null)
			{
				if(akeeba_error_callback != null)
				{
					akeeba_error_callback(message, uuid);
				}
			}
			else
			{
				errorCallback(message, uuid);
			}
		}
	};
	$.ajax( structure );
};

AkeebaBackupAPI.prototype.getURL = function(method, data)
{
	// Create a random salt string
	var chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXTZabcdefghiklmnopqrstuvwxyz";
	var string_length = 64;
	var salt = '';
	for (var i=0; i<string_length; i++) {
		var rnum = Math.floor(Math.random() * chars.length);
		salt += chars.substring(rnum,rnum+1);
	}
	
	var challengeString = salt + this.secret;
	var md5Challenge = md5(challengeString);
	var challenge = salt+':'+md5Challenge;
	
	var abReqBody = {
			method: method,
			data: data,
			challenge: challenge
	};
	
	var bodyJSON = json_encode(abReqBody);
	
	var abReqObj = {
		'encapsulation': 1,
		'body': bodyJSON
	};
	
	var requestjson = json_encode(abReqObj);
	var myURL = rtrim(this.URL,'/') + '/index.php?option=com_akeeba&view=json&format=raw&json=' + urlencode(requestjson);
	
	return myURL;
};

/**
 * Runs an Akeeba Backup remote API method and returns its results
 * @param method String The method to run
 * @param data Object The raw data to send as method parameters
 * @param callback function The function to call upon return of AJAX data (or on error, it's the same thing!)
 * @return Object
 */
AkeebaBackupAPI.prototype.runMethod = function(method,data,callback,timeout)
{
	// Use a sensible timeout
	if(empty(timeout)) {
		timeout = 600000;
	}
	
	var myURL = this.getURL(method, data);

	var akeebaAPI = this;
	
	this.doAJAX(
		myURL,
		function(msg, uuid){
			var abResObj = json_decode(msg);
			var abResBody = msg.body;
			
			var result = {
				status: abResBody.status,
				msg: json_decode(abResBody.data)
			};

			callback(result, uuid);
		},
		function(error, uuid){
			var result = {
				status: 600,
				msg: error,
				encapsulation: 1
			};
			
			callback(result, uuid);
		},
		timeout
	);
};

/**
 * Returns true if this site is the only selected site in the interface
 * @return
 */
AkeebaBackupAPI.prototype.isCurrent = function(foobar) {
	if(count(akeeba_remote_selected_sites) > 1) return false;
	if(akeeba_remote_selected_sites[0] == this.uuid) return true;
	return false;
};

/**
 * Gets the Akeeba Backup version information and updates the caches and the GUI
 * @param usecache Boolean True (default) to use the cached data
 * @return
 */
AkeebaBackupAPI.prototype.getVersion = function(usecache)
{
	this.addLog('Running getVersion');
	
	if(usecache == null) usecache = true;
	if(usecache == true) {
		if(this.versionCache != null) {
			this.addLog('Using cached version data for getVersion');
			this.getVersion_cb({
				status: 200,
				msg: this.versionCache
			}, this.uuid);
			return;
		}
	}
	
	this.runMethod('getVersion', {}, this.getVersion_cb);
};

/**
 * Callback method for getVersion
 * @param data
 * @return
 */
AkeebaBackupAPI.prototype.getVersion_cb = function(data, uuid) {
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	if(data.status != 200) {
		// An error occurred; get the error message
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('getVersion failed with an error:');
		akeebaAPI.addLog(akeebaAPI.errorMessage);
		// Update the site status
		siteManager.setSiteStatus(uuid, 'genericerror');
		akeebaAPI.versionCache = null;
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
	} else {
		// Cache the data
		akeebaAPI.addLog('getVersion: Got data');
		akeebaAPI.versionCache = data.msg;
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) {
			// If the site is currently selected, we also have to fetch the backup profiles array
			akeebaAPI.getProfiles();
		} else if(akeebaAPI.startBackupAfterGetProfiles) {
			akeebaAPI.startBackupAfterGetProfiles = false;
			startCurrentBackup(uuid);
		}
	}
};

/**
 * Gets the Akeeba Backup profiles list and caches it
 * @param usecache Boolean True (default) to use the cached data
 * @return
 */
AkeebaBackupAPI.prototype.getProfiles = function(usecache)
{
	this.addLog('Running getProfiles');
	
	if(usecache == null) usecache = true;
	if(usecache == true) {
		if(this.profilesCache != null) {
			this.addLog('Using cached data for getProfiles');
			
			this.getProfiles_cb({
				status: 200,
				msg: this.profilesCache
			}, this.uuid);
			return;
		}
	}
	
	this.runMethod('getProfiles', {}, this.getProfiles_cb, 7000);
};

AkeebaBackupAPI.prototype.getProfiles_cb = function(data, uuid) {
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	if(data.status != 200) {
		// An error occurred; get the error message
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('getProfiles failed with an error:');
		akeebaAPI.addLog(akeebaAPI.errorMessage);
		
		// Update the site status
		siteManager.setSiteStatus(uuid, 'genericerror');
		akeebaAPI.versionCache = null;
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
	} else {
		akeebaAPI.addLog('getProfile: Got profiles data');
		
		// Cache the data
		akeebaAPI.profilesCache = data.msg;
		// Create an indexed array and store it alongside the record
		if(!empty(akeebaAPI.profilesCache)) {
			var profileData = [];
			$.each(akeebaAPI.profilesCache,function(index, item){
				profileData[item.id] = item.name;
				akeebaAPI.addLog('-- Profile #'+item.id+': '+item.name);
			});
			$(siteSelector).data('akeeba_profiles', profileData);			
		}
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) {
			renderSiteVersionPane();
		}
		if(akeebaAPI.startBackupAfterGetProfiles) {
			akeebaAPI.startBackupAfterGetProfiles = false;
			akeebaAPI.startBackup();
		}
	}
};


/**
 * Start backing up the website
 */
AkeebaBackupAPI.prototype.startBackup = function(backupID)
{
	// If we have not yet retrieved the version information, do it now
	if(empty(this.versionCache)) {
		this.startBackupAfterGetProfiles = true;
		this.getVersion(false);
		return;
	}
	
	if(this.versionCache.api < this.minimumAPIVersion) {
		// The API level is too low. Fail immediately.
		this.errorMessage = Trans.t('NEWERVERSIONREQUIRED');
		siteManager.setSiteStatus(this.uuid, 'error');
		if(this.isCurrent()) renderSingleSitePane();
		
		// Also create a desktop notification
		var notification = Titanium.Notification.createNotification(window);
		notification.setTitle(Trans.t('BACKUPCANTPROCEED'));
		notification.setMessage(this.errorMessage);
		notification.setIcon('images/fail.png');
		notification.setTimeout(10);
		notification.show();
		
		return;

	}
	
	// Initialize the backup ID (null = we got no such data yet)
	this.backupID = null;
	// First, the GUI
	this.backupInfo = {
		'Domain': 'init',
		'Step': '',
		'Substep': '',
		'Warnings': []
	};
	siteManager.setSiteStatus(this.uuid, 'backup');
	if(this.isCurrent()) renderSingleSitePane();
	// Start the backup. As simple as it gets, right?
	this.addLog('Running startBackup');
	this.runMethod('startBackup', {profile: this.profile}, this.stepBackup);
};

/**
 * Steps through the backup process. This method acts as a callback to startBackup() and
 * as a callback to itself.
 * @param data Object The backup notification message.
 * @param uuid String The UUID of the site we're backing up.
 */
AkeebaBackupAPI.prototype.stepBackup = function(data, uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	var hasError = false;
	var errorType = 'genericerror';
	
	// If the status indicates a comm error, let's error out
	if(data.status != 200) {
		hasError = true;
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('Backup failed (hard error): ' + data.msg);
	} else {
		// Do we have a "soft" error?
		akeebaAPI.backupInfo = data.msg;
		if(data.msg.Error) {
			hasError = true;
			errorType = 'error';
			akeebaAPI.errorMessage = data.msg.Error;
			akeebaAPI.addLog('Backup failed (soft error): ' + data.msg.Error);
		}
	}
	
	// Error handling, if required
	if(hasError) {
		siteManager.setSiteStatus(uuid, errorType);
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
		
		// Create a desktop notification
		var notification = Titanium.Notification.createNotification(window);
		notification.setTitle(Trans.t('BACKUPFAILED'));
		notification.setMessage(
				sprintf(Trans.t('BACKUPERRORMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
		);
		notification.setIcon('images/fail.png');
		notification.setTimeout(10);
		notification.show();
		
		return;
	}
	
	// No error, fetch the backup ID
	if(akeebaAPI.backupInfo.BackupID)
	{
		akeebaAPI.backupID = akeebaAPI.backupInfo.BackupID;
	}
	
	// If no error occurred, check if we're done
	if(akeebaAPI.backupInfo.HasRun && !akeebaAPI.cancelBackup) {
		akeebaAPI.addLog('stepBackup: more steps required');
		// More steps required, so update GUI...
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
		// ...and run the next step
		akeebaAPI.runMethod('stepBackup', {}, akeebaAPI.stepBackup);
	} else if(akeebaAPI.cancelBackup) {
		akeebaAPI.addLog('stepBackup: user cancelled the backup job');
		// Backup cancelled
		akeebaAPI.cancelBackup = false;
		siteManager.setSiteStatus(uuid, 'idle');
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
	} else {
		// Backup finished
		akeebaAPI.addLog('stepBackup: just finished');
		akeebaAPI.finishedBackup(data, uuid);
	}
};

/**
 * Post-backup callback. It is called by stepBackup when the backup is over (but not when
 * the backup fails). It is supposed to update the GUI and start the download process if
 * user's preferences dictate so.
 * @param data Object The backup notification message
 * @param uuid String The UUID of the site we're backing up
 * @return
 */
AkeebaBackupAPI.prototype.finishedBackup = function(data, uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	// Set the site's state as 'idle'
	siteManager.setSiteStatus(uuid, 'idle');
	//$(siteSelector).data('akeeba_api_object', akeebaAPI);
	if(akeebaAPI.isCurrent()) renderSingleSitePane();
	
	// Create a desktop notification
	var notification = Titanium.Notification.createNotification(window);
	notification.setTitle(Trans.t('BACKUPFINISHED'));
	notification.setMessage(
			sprintf(Trans.t('BACKUPFINISHEDMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
	);
	notification.setTimeout(10);
	notification.setIcon('images/success.png');
	notification.show();
	
	// Save the last successful backup date
	var siteDef = siteManager.getSiteDefinition(uuid);
	siteDef.lastSuccessfulBackup = time();
	siteDef.lastSuccessfulFilename = '';
	siteManager.setSiteDefinition(uuid, siteDef);
	
	// Check if the user wishes to download the backup files
	if(akeebaAPI.download) {
		akeebaAPI.startDownload(akeebaAPI.backupID);
	} else {
		saveAppProperties();
	}
};

/**
 * Begins downloading all parts of a backup archive belonging to the backup attempt with
 * record ID of backupid
 * @param backupid Integer The backup record ID to download
 */
AkeebaBackupAPI.prototype.startDownload = function(backupid, allowdelete)
{
	siteManager.setSiteStatus(this.uuid, 'download');
	if(this.isCurrent()) {
		renderSingleSitePane();
		renderSiteDownloadPane();
	}
	
	// Parse the allowdelete parameter
	if(empty(allowdelete)) {
		allowdelete = true;
	}
	if(allowdelete) {
		// We are allowed to delete. Does the user wish to do so?
		this.DownloadDeleteAfter = this.deleteAfterDownload;
	} else {
		// We are not allowed to delete.
		this.DownloadDeleteAfter = false;
	}
	
	// Check for cached data
	if(this.DownloadBackupID == backupid) {
		this.startDownload_cb({status: 200, msg:this.DownloadBackupInformation}, this.uuid);
	} else {
		this.DownloadBackupID = backupid;
		this.DownloadBackupInformation = null;
		this.DownloadTotalSize = null;
		this.DownloadDownloadedSize = null;
		this.DownloadQueue = null;
		
		this.runMethod('getBackupInfo', {backup_id: backupid}, this.startDownload_cb);
	}
};

/**
 * Callback function being called when starting a download. It sets up the download
 * and seeds the data used for the progress display.
 * @param data
 * @param uuid
 */
AkeebaBackupAPI.prototype.startDownload_cb = function(data, uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	if(data.status != 200) {
		// An error occurred; get the error message
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('getBackupInfo failed with an error:');
		akeebaAPI.addLog(akeebaAPI.errorMessage);
		
		// Update the site status
		siteManager.setSiteStatus(uuid, 'dlerror');
		akeebaAPI.DownloadBackupID = null;
		akeebaAPI.DownloadBackupInformation = null;
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
		
		var notification = Titanium.Notification.createNotification(window);
		notification.setTitle(Trans.t('DOWNLOADFAILED'));
		notification.setMessage(
				sprintf(Trans.t('DOWNLOADFAILEDMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
		);
		notification.setIcon('images/fail.png');
		notification.setTimeout(10);
		notification.show();
		
		return;
	} else {
		akeebaAPI.addLog('getBackupInfo: Got backup record information');
		
		// Cache the data
		akeebaAPI.DownloadBackupInformation = data.msg;
		
		// Do we have an obsolete, error or otherwise non-OK record?
		if(!akeebaAPI.DownloadBackupInformation.filesexist || empty(akeebaAPI.DownloadBackupInformation.filenames)) {
			akeebaAPI.errorMessage = Trans.t('NOBACKUPFILESFORRECORD');
			akeebaAPI.addLog(akeebaAPI.errorMessage);
			// Update the site status
			siteManager.setSiteStatus(uuid, 'dlerror');
			akeebaAPI.DownloadBackupID = null;
			akeebaAPI.DownloadBackupInformation = null;
			//$(siteSelector).data('akeeba_api_object', akeebaAPI);
			if(akeebaAPI.isCurrent()) renderSingleSitePane();
			
			var notification = Titanium.Notification.createNotification(window);
			notification.setTitle(Trans.t('DOWNLOADFAILED'));
			notification.setMessage(
					sprintf(Trans.t('DOWNLOADFAILEDMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
			);
			notification.setIcon('images/fail.png');
			notification.setTimeout(10);
			notification.show();
			
			return;
		}
		
		// Calculate total size and seed the download queue
		if(!empty(akeebaAPI.DownloadBackupInformation.filenames)) {
			
			// Save the last successful backup file name
			var siteDef = siteManager.getSiteDefinition(uuid);
			siteDef.lastSuccessfulFilename = akeebaAPI.DownloadBackupInformation.filenames[0].name;
			siteManager.setSiteDefinition(uuid, siteDef);
			saveAppProperties();
			
			var backupFragments = [];
			akeebaAPI.DownloadTotalSize = 0;
			akeebaAPI.DownloadDownloadedSize = 0;
			$.each(akeebaAPI.DownloadBackupInformation.filenames, function(index,item){
				var part = item.part;
				var name = item.name;
				var fragments = parseInt(item.size / 1048756);
				if(fragments * 1048756 < item.size) fragments++;
				for(i = 0; i < fragments; i ++) {
					var newFragment = {
						backup_id: akeebaAPI.DownloadBackupID,
						part_id: item.part,
						segment: i + 1,
						name: item.name,
						isFirst: (i == 0)
					};
					array_push(backupFragments, newFragment);
				}
				
				akeebaAPI.DownloadTotalSize += item.size;
			});
			
			// Store the download queue
			akeebaAPI.DownloadQueue = backupFragments;
		} else {
			akeebaAPI.DownloadQueue = [];
		}
		
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		// Refresh GUI
		siteManager.setSiteStatus(uuid, 'download');
		if(akeebaAPI.isCurrent()) {
			renderSingleSitePane();
			renderSiteDownloadPane();
		}
		
		// Start downloading
		akeebaAPI.downloadSegment(uuid);
	}	
};

/**
 * Downloads the next segment in the queue
 * @param uuid
 * @return
 */
AkeebaBackupAPI.prototype.downloadSegment = function(uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	

	if(count(akeebaAPI.DownloadQueue)) {
		// Get the next segment to download
		var nextSegment = array_shift(akeebaAPI.DownloadQueue);
		// Calculate the absolute file path
		akeebaAPI.DownloadOutputFilename = Titanium.Filesystem.getFile(akeebaAPI.downloadDirectory, nextSegment.name);
		// If we're starting a new file, remove any existing ones
		if(nextSegment.isFirst) akeebaAPI.DownloadOutputFilename.deleteFile();
		// Save the new state -- Do I really need that?
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		// Then run the download remote API method to fetch the raw file data
		akeebaAPI.nextSegment = nextSegment;
		akeebaAPI.downloadRetries = 0;
		akeebaAPI.runMethod('download', nextSegment, akeebaAPI.downloadSegment_cb);
	} else {
		// Should we delete the files afterwards?
		if(akeebaAPI.DownloadDeleteAfter) {
			// Delete the files
			akeebaAPI.deleteFiles(akeebaAPI.DownloadBackupID, uuid);
		} else {
			// No need to delete the files afterwards; mark ourselves as complete
			akeebaAPI.downloadIsComplete(uuid);
		}		
	}
};

/**
 * Retries download
 * @param uuid
 * @return
 */
AkeebaBackupAPI.prototype.retryDownload = function(uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	// Increase the retries count
	akeebaAPI.downloadRetries++;
	// Run the download, again!
	akeebaAPI.runMethod('download', akeebaAPI.nextSegment, akeebaAPI.downloadSegment_cb);
};

/**
 * Saves the downloaded data
 * @param data
 * @param uuid
 * @return
 */
AkeebaBackupAPI.prototype.downloadSegment_cb = function(data,uuid)
{
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	if(data.status != 200) {
		// How many retries do I have?
		if(akeebaAPI.downloadRetries < 3) {
			// Give it a cool off period of 3 - 9 seconds and retry
			setTimeout("$('#site-"+uuid+"').data('akeeba_api_object').retryDownload('"+uuid+"')", (akeebaAPI.downloadRetries + 1) * 3000);
			return;
		}
		
		// An error occurred and we've already tried 3 times; get the error message
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('download failed with an error:');
		akeebaAPI.addLog(akeebaAPI.errorMessage);
		
		// Update the site status
		siteManager.setSiteStatus(uuid, 'dlerror');
		akeebaAPI.DownloadBackupID = null;
		akeebaAPI.DownloadBackupInformation = null;
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
		
		var notification = Titanium.Notification.createNotification(window);
		notification.setTitle(Trans.t('DOWNLOADFAILED'));
		notification.setMessage(
				sprintf(Trans.t('DOWNLOADFAILEDMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
		);
		notification.setIcon('images/fail.png');
		notification.setTimeout(10);
		notification.show();

		return;
	} else {
		// Decode the binary data and write it to file
		var fileName = akeebaAPI.DownloadOutputFilename.toString();
		var encodedData = data.msg;
		Titanium.API.debug('Writing '+encodedData.length+' bytes of Base64 data to '+fileName);
		var dataSize = writeit(encodedData, fileName);
		Titanium.API.debug('Written '+dataSize+' raw bytes');

		akeebaAPI.DownloadDownloadedSize += dataSize;
		// Update GUI (sidebar & download status)
		siteManager.setSiteStatus(uuid,'download');
		//$(siteSelector).data('akeeba_api_object', akeebaAPI);
		if(akeebaAPI.isCurrent()) {
			renderSingleSitePane();
			renderSiteDownloadPane();
		}
		// Download the next segment
		akeebaAPI.downloadSegment(uuid);			
	}
};

/**
 * Notifies the user that the download of the files is now complete
 */
AkeebaBackupAPI.prototype.downloadIsComplete = function(uuid) {
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	

	// Set the site's state as 'idle'
	siteManager.setSiteStatus(uuid, 'idle');
	if(akeebaAPI.isCurrent()) renderSingleSitePane();
	
	// Create a desktop notification
	var notification = Titanium.Notification.createNotification(window);
	notification.setTitle(Trans.t('DOWNLOADCOMPLETE'));
	notification.setMessage(
			sprintf(Trans.t('DOWNLOADCOMPLETEMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
	);
	notification.setTimeout(10);
	notification.setIcon('images/success.png');
	notification.show();
};

/**
 * Deletes the files of a backup record
 * @param downloadID int The backup record to delete its files
 * @param uuid string The UUID of the site whose backup record we're referencing
 * @return
 */
AkeebaBackupAPI.prototype.deleteFiles = function(downloadID, uuid) {
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	
	
	var data = {
		'backup_id': downloadID
	};
	
	akeebaAPI.addLog('Deleting files of backup record '+downloadID);
	
	akeebaAPI.runMethod('deleteFiles', data, akeebaAPI.deleteFiles_cb);
};

AkeebaBackupAPI.prototype.deleteFiles_cb = function(data, uuid) {
	var siteSelector = '#site-' + uuid;
	var akeebaAPI = $(siteSelector).data('akeeba_api_object');	

	if(data.status != 200) {
		// An error occurred; get the error message
		if(empty(data.msg)) {
			akeebaAPI.errorMessage = akeebaAPI.errormessages[data.status];
		} else {
			akeebaAPI.errorMessage = data.msg;
		}
		akeebaAPI.addLog('Deleting archive files failed with an error:');
		akeebaAPI.addLog(akeebaAPI.errorMessage);
		
		// Update the site status
		siteManager.setSiteStatus(uuid, 'dlerror');
		akeebaAPI.DownloadBackupID = null;
		akeebaAPI.DownloadBackupInformation = null;
		if(akeebaAPI.isCurrent()) renderSingleSitePane();
		
		var notification = Titanium.Notification.createNotification(window);
		notification.setTitle(Trans.t('DELETEERROR'));
		notification.setMessage(
				sprintf(Trans.t('DELETEERRORMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
		);
		notification.setIcon('images/fail.png');
		notification.setTimeout(10);
		notification.show();
	} else {
		if(akeebaAPI.DownloadDeleteAfter) {
			// We were automatically called during backup. Mark downloading the backup complete.
			akeebaAPI.downloadIsComplete(uuid);
			// Create a desktop notification
			var notification = Titanium.Notification.createNotification(window);
			notification.setTitle(Trans.t('DELETEOK'));
			notification.setMessage(
					sprintf(Trans.t('DELETEOKMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
			);
			notification.setTimeout(10);
			notification.setIcon('images/success.png');
			notification.show();			
		} else {
			// Set the site's state as 'idle'
			siteManager.setSiteStatus(uuid, 'idle');
			if(akeebaAPI.isCurrent()) renderSingleSitePane();
			
			// Create a desktop notification
			var notification = Titanium.Notification.createNotification(window);
			notification.setTitle(Trans.t('DELETEOK'));
			notification.setMessage(
					sprintf(Trans.t('DELETEOKMESSAGE'), Titanium.API.Application.getName(), $(siteSelector+' span.siteName:first').text())
			);
			notification.setTimeout(10);
			notification.setIcon('images/success.png');
			notification.show();			
		}
	}
};

/**
 * Enqueue the site
 * TODO
 */
AkeebaBackupAPI.prototype.enqueue = function()
{
	if(this.cancelBackup) return;
	
	// Initialize the backup ID (null = we got no such data yet)
	this.backupID = null;
	// First, the GUI
	this.backupInfo = {
		'Domain': 'queue',
		'Step': '',
		'Substep': '',
		'Warnings': []
	};
	siteManager.setSiteStatus(this.uuid, 'queue');
	if(this.isCurrent()) renderSingleSitePane();
	setTimeout("startCurrentBackup('"+this.uuid+"');",1000);
};