/**
 * Akeeba Remote Control
 * A utility to remotely control Akeeba Backup and Admin Tools
 * 
 * @version $Id$
 * @copyright 2008-2010 Nicholas K. Dionysopoulos / AkeebaBackup.com
 * @license GNU/GPL v3 or later
 */

/**
 * Site manager prototype
 */
function AkeebaSiteManager()
{
	this.siteDefinitions = {};
	this.siteStatuses = {};
	
	this.__defineGetter__("sites",function(){
		return sites.map(this.getSiteDefinition(uuid));
	});
	this.__defineSetter__("sites",function(){
		return sites.map(this.setSiteDefinition(uuid));
	});
}

AkeebaSiteManager.prototype.siteDefinitions = {};

AkeebaSiteManager.prototype.siteStatuses = {};

/**
 * Adds a site to the Site Manager (and the associated display!)
 * @param uuid String A UUID to uniquely identify the site in the manager
 * @param siteDef Object The site definition
 * @return
 */
AkeebaSiteManager.prototype.addSite = function(uuid, siteDef)
{
	// Does the UUID already exist?
	if( array_key_exists(uuid, this.siteDefinitions) ) return false;
	
	// Create new entries
	this.siteDefinitions[uuid] = {};
	this.siteStatuses[uuid] = {};
	
	// Create new list item
	var myID = 'site-' + uuid;
	var newItem = $(document.createElement('li'))
		.attr('id',myID)
		.addClass('ui-widget-content')
		.appendTo('#siteList');
	$(document.createElement('span')).addClass('handle').append(
			$(document.createElement('span')).addClass('ui-icon').addClass('ui-icon-arrowthick-2-n-s')
	).appendTo(newItem);
	$(document.createElement('span')).addClass('siteName').appendTo(newItem);
	$(document.createElement('span')).addClass('ui-icon').addClass('siteIcon').appendTo(newItem);
	$(document.createElement('span')).addClass('siteStatus').appendTo(newItem);
	
	// Update internal data and display
	this.setSiteDefinition(uuid, siteDef);
	this.setSiteStatus(uuid, 'idle');
	
	// Create an AkeebaBackupAPI object and store it in the newItem's data
	var APIObject = new AkeebaBackupAPI();
	APIObject.loadSite(uuid);
	newItem.data('akeeba_api_object', APIObject);
};

/**
 * Removes a site from the Site Manager
 * @param uuid String The UUID of the site to remove
 * @return Boolean True if the UUID existed and we could remove the site
 */
AkeebaSiteManager.prototype.removeSite = function(uuid)
{
	// Does the UUID exist?
	if( !array_key_exists(uuid, this.siteDefinitions) ) return false;
	
	// Is the site in any state except backup or download?
	if( in_array(this.siteStatuses[uuid], ['backup','download'] ) ) return false;
	
	// Remove the list item
	$('#site-'+uuid).remove();
	// Remove the array keys
	delete this.siteDefinitions[uuid];
	delete this.siteStatuses[uuid];

};

/**
 * Returns the site definition of a given site
 * @param uuid String The UUID of the site
 * @return Object
 */
AkeebaSiteManager.prototype.getSiteDefinition = function(uuid)
{
	if( !array_key_exists(uuid, this.siteDefinitions) ) return false;

	return this.siteDefinitions[uuid];
};

/**
 * Sets or updates the site's definition
 * @param uuid String The UUID of the site
 * @param siteDef Object The site definition parameters
 * @return null
 */
AkeebaSiteManager.prototype.setSiteDefinition = function(uuid, siteDef)
{
	// Does the UUID already exist?
	if( !array_key_exists(uuid, this.siteDefinitions) ) return false;
	
	// Store the definition
	this.siteDefinitions[uuid] = siteDef;
	// Set the site name label
	$('#site-'+uuid+' span.siteName').first().text(siteDef.name);
};

AkeebaSiteManager.prototype.getNumSitesWithStatus = function(someStatus)
{
	var counter = 0;
	for(var key in this.siteStatuses)
	{
		if(this.siteStatuses[key] == someStatus) counter++;
	}
	
	return counter;
}

/**
 * Returns the status of a given site
 * @param uuid String The UUID of the site
 * @return String
 */
AkeebaSiteManager.prototype.getSiteStatus = function(uuid)
{
	if( !array_key_exists(uuid, this.siteDefinitions) ) return false;

	return this.siteStatuses[uuid];
};

/**
 * Sets or updates the site's status (and the display)
 * @param uuid String The UUID of the site
 * @param status String
 */
AkeebaSiteManager.prototype.setSiteStatus = function(uuid, status)
{
	// Does the UUID already exist?
	if( !array_key_exists(uuid, this.siteDefinitions) ) return false;
	
	var newStatus = status;
	var newIcon = 'ui-icon-bullet';
	var newText = Trans.t('STATUSIDLE');
	var errorState = false;
	
	switch(status) {
		case 'queue':
			newIcon = 'ui-icon-clock';
			newText = Trans.t('STATUSQUEUED');
			break;
	
		case 'backup':
			newIcon = 'ui-icon-circle-triangle-e';
			newText = Trans.t('STATUSBACKUP');
			break;

		case 'download':
			// Get the API object
			var siteSelector = '#site-' + uuid;
			var akeebaAPI = $(siteSelector).data('akeeba_api_object');
			var percentage = 0;
			
			newIcon = 'ui-icon-circle-arrow-s';
			
			if( (akeebaAPI.DownloadTotalSize > 0) && (akeebaAPI.DownloadDownloadedSize > 0) ) {
				percentage = 100 * akeebaAPI.DownloadDownloadedSize / akeebaAPI.DownloadTotalSize;
				var percentage = parseInt(percentage);
			}
			
			newText = sprintf(Trans.t('STATUSDOWNLOAD'),percentage);
			break;

		case 'dlerror':
			newIcon = 'ui-icon-alert';
			newText = Trans.t('STATUSDLERROR');
			errorState = true;
			break;

		case 'error':
			newIcon = 'ui-icon-circle-close';
			newText = Trans.t('STATUSERROR');
			errorState = true;
			break;
			
		case 'genericerror':
			newIcon = 'ui-icon-circle-close';
			newText = Trans.t('STATUSGENERICERROR');
			errorState = true;
			break;

		case 'idle':
		default:
			newStatus = 'idle';
			break;
	}
	
	// Store the status
	this.siteStatuses[uuid] = newStatus;
	// Set/unset the error class
	if(errorState) {
		$('#site-'+uuid).addClass('ui-state-error');
	} else {
		$('#site-'+uuid).removeClass('ui-state-error');
	}
	// Set the status icon
	$('#site-'+uuid+' span.siteIcon').attr('class','ui-icon siteIcon '+newIcon);
	// Set the status label
	$('#site-'+uuid+' span.siteStatus').text(newText);
};

/**
 * Serializes the site information to a JSON object
 * @return string
 */
AkeebaSiteManager.prototype.getSerialized = function()
{
	var orderedDefinitions = {};
	
	$('#siteList li').each(function(idx, el){
		var siteID = $(el).attr('id');
		var uuid = substr(siteID, 5);
		var sitedef = siteManager.getSiteDefinition(uuid);
		orderedDefinitions[uuid] = sitedef;
	});
	
	return json_encode(orderedDefinitions);
};

/**
 * @var AkeebaSiteManager The global instance of the Site Manager object
 */
var siteManager = new AkeebaSiteManager();