<?php
/**
 * Akeeba Engine
 * The modular PHP5 site backup engine
 * @copyright Copyright (c)2009-2011 Nicholas K. Dionysopoulos
 * @license GNU GPL version 3 or, at your option, any later version
 * @package akeebaengine
 * @version $Id$
 *
 * Contains: Dropbox Uploader 1.1.5 - Copyright (c) 2009 Jaka Jancar
 *
 * As per its license, we are obliged to include the following notice:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

// Protection against direct access
defined('AKEEBAENGINE') or die('Restricted access');

class AEPostprocDropbox extends AEAbstractPostproc
{
	/** @var string DropBox login email */
    protected $email;
    /** @var string DropBox password */
    protected $password;
    /** @var bool Are we logged in DropBox yet? */
    protected $loggedIn = false;
    /** @var string DropBox cookies */
    protected $cookies = array();
    
    /** @var int The retry count of this file (allow up to 2 retries after the first upload failure) */
    private $tryCount = 0;
	
	public function __construct()
	{
		parent::__construct();
		
		$this->can_download_to_browser = true;
		$this->can_delete = true;
		/*
		$this->can_download_to_file = true;
		*/
	}

	public function processPart($absolute_filename)
	{
		$settings = $this->_getSettings();
		if($settings === false) return false;
		extract($settings);

		// Connect and send
		$result = $this->dropbox_init($email, $password);
		if($result === false)
		{
			return false;
		}
		
		// Store the absolute remote path in the class property
		$this->remote_path = $directory.'/'.basename($absolute_filename);

		$result = $this->dropbox_upload(
			$absolute_filename,											// File to read from
			$directory													// Remote directory
		);

		if($result === false) {
			// If it is a hard error, fail immediately
			if($this->getError()) return false;
			// If it is a soft error (e.g. connection issues) let's retry
			$this->tryCount++;
			// However, if we've already retried twice, we stop retrying and call it a failure
			if($this->tryCount > 2) return false;
			return -1;
		} else {
			// Upload complete. Reset the retry counter.
			$this->tryCount = 0;
			return true;
		}
	}
	
	public function downloadToFile($remotePath, $localFile, $fromOffset = null, $length = null)
	{
		// Get settings
		$settings = $this->_getSettings();
		if($settings === false) return false;
		extract($settings);
		
		// Get the remote path's components
		$remoteDirectory = trim(dirname($remotePath),'/');
		$remoteFilename = basename($remotePath);
		
		// Connect to DropBox
		$result = $this->dropbox_init($email, $password);
		if($result === false)
		{
			return false;
		}
		
		// Get the list of files in the remote directory
		$contents = $this->dropbox_list($remoteDirectory);
		if(!array_key_exists($remoteFilename, $contents->files)) {
			$this->setError("Remote file $remotePath not found");
			return false;
		}
		
		$url = $contents->files[$remoteFilename]->url;
		
		return $this->dropbox_download($url, $localFile, $fromOffset, $length);
	}
	
	public function delete($path)
	{
		// Get settings
		$settings = $this->_getSettings();
		if($settings === false) return false;
		extract($settings);
		
		// Connect to DropBox
		$result = $this->dropbox_init($email, $password);
		if($result === false)
		{
			return false;
		}

		// Get the token
		$data = $dropbox->request('https://www.dropbox.com/home', false, array(), false);
		preg_match('#\s{2}TOKEN:\s\'(([a-z,0-9]){8,16})\',#', $data, $m);
		$token = $m[1];

		$postData = array(
			'files'		=> '/'.trim($path,'/'),
			't'			=> $token
		);
		$data = $dropbox->request('https://www.dropbox.com/cmd/delete?long_running=1', true, $postData);

		if(strpos($data, 'HTTP/1.1 200')) {
			return true;
		} else {
			return false;
		}
	}
	
	protected function _getSettings()
	{
		// Retrieve engine configuration data
		$config =& AEFactory::getConfiguration();

		$email		= trim( $config->get('engine.postproc.dropbox.email', '') );
		$password	= trim( $config->get('engine.postproc.dropbox.password', '') );
		$directory	= $config->get('volatile.postproc.directory', null);
		if(empty($directory)) $directory	= $config->get('engine.postproc.dropbox.directory', '');

		// Sanity checks
		if(empty($email))
		{
			$this->setError('You have not set up your DropBox email');
			return false;
		}

		if(empty($password))
		{
			$this->setError('You have not set up your DropBox password');
			return false;
		}

		// Fix the directory name, if required
		if(!empty($directory))
		{
			$directory = trim($directory);
			$directory = ltrim( AEUtilFilesystem::TranslateWinPath( $directory ) ,'/');
		}
		else
		{
			$directory = '';
		}
		
		// Parse tags
		$directory = AEUtilFilesystem::replace_archive_name_variables($directory);
		$config->set('volatile.postproc.directory', $directory);

		return array(
			'email'		=> $email,
			'password'	=> $password,
			'directory'	=> $directory
		);
	}

    /**
     * Initializes the DropBox connectivity
     *
     * @param string $email
     * @param string|null $password
     */
    protected function dropbox_init($email, $password) {
        if (!extension_loaded('curl'))
        {
        	$this->setError('DropboxUploader requires the cURL extension.');
        	return false;
        }
        $this->email = $email;
        $this->password = $password;
    }

    protected function dropbox_upload($filename, $remoteDir='/') {
        if (!file_exists($filename) or !is_file($filename) or !is_readable($filename))
        {
        	$this->setError("File '$filename' does not exist or is not readable.");
        	return false;
        }

        if (!is_string($remoteDir))
        {
            $this->setError("Remote directory must be a string, is ".gettype($remoteDir)." instead.");
            return false;
        }

        if (preg_match("/.+\.\..+/",$remoteDir))
        {
			$this->setError("Remote directory is impossible");
            return false;
        }

        if (!$this->loggedIn)
        {
            $status = $this->dropbox_login();
            if(!$status) return false;
        }

        $data = $this->dropbox_request('https://www.dropbox.com/home');
        if($data === false) return false;
        $token = $this->dropbox_extractToken($data, 'https://dl-web.dropbox.com/upload');
		if($token === false) return false;

        $data = $this->dropbox_request('https://dl-web.dropbox.com/upload', true, array('plain'=>'yes', 'file'=>'@'.$filename, 'dest'=>$remoteDir, 't'=>$token));
        if($data === false) return false;
        if (strpos($data, 'HTTP/1.1 302 FOUND') === false)
        {
			$this->setWarning('Upload failed!');
			return false;
        }

        return true;
    }

    protected function dropbox_login() {
        $data = $this->dropbox_request('https://www.dropbox.com/login');
        if($data === false) return false;
        $token = $this->dropbox_extractToken($data, '/login');
        if($token === false) return false;

        $data = $this->dropbox_request('https://www.dropbox.com/login', true, array('login_email'=>$this->email, 'login_password'=>$this->password, 't'=>$token));

        if (stripos($data, 'location: /home') === false)
        {
            $this->setWarning('DropBox login unsuccessful.');
            return false;
        }

        $this->loggedIn = true;
        return true;
    }

    protected function dropbox_request($url, $post=false, $postData=array(), $noHeader = false) {
        $ch = curl_init();
		@curl_setopt($ch, CURLOPT_CAINFO, AKEEBA_CACERT_PEM);
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        if($noHeader) {
			curl_setopt($ch, CURLOPT_HEADER, 0);
			@curl_setopt($ch, CURLOPT_FOLLOWLOCATION, 1);
			@curl_setopt($ch, CURLOPT_MAXREDIRS, 100);
		} else {
			curl_setopt($ch, CURLOPT_HEADER, 1);
		}
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        if ($post) {
            curl_setopt($ch, CURLOPT_POST, $post);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $postData);
        }

        // Send cookies
        $rawCookies = array();
        foreach ($this->cookies as $k=>$v)
            $rawCookies[] = "$k=$v";
        $rawCookies = implode(';', $rawCookies);
        curl_setopt($ch, CURLOPT_COOKIE, $rawCookies);

        $data = curl_exec($ch);

        if ($data === false)
        {
			$this->setWarning('Cannot execute request: '.curl_error($ch));
			return false;
        }

        // Store received cookies
        preg_match_all('/Set-Cookie: ([^=]+)=(.*?);/i', $data, $matches, PREG_SET_ORDER);
        foreach ($matches as $match)
            $this->cookies[$match[1]] = $match[2];

        curl_close($ch);

        return $data;
    }

    protected function dropbox_extractToken($html, $formAction) {
        if (!preg_match('/<form [^>]*'.preg_quote($formAction, '/').'[^>]*>.*?(<input [^>]*name="t" [^>]*value="(.*?)"[^>]*>).*?<\/form>/is', $html, $matches) || !isset($matches[2]))
        {
			$this->setWarning("Cannot extract token! (form action=$formAction)");
			return false;
        }
        return $matches[2];
    }

	/**
	 * List the contents of a directory
	 * @param type $path 
	 */
	protected function dropbox_list($path)
	{
		$directory = trim(dirname($path), '/');
		
		if(!$this->loggedIn) {
			$this->dropbox_login();
		}
		
		// Extract the path listing token
		$data = $this->dropbox_request('https://www.dropbox.com/home', false, array(), false);
		preg_match('#\s{2}TOKEN:\s\'(([a-z,0-9]){8,16})\',#', $data, $m);
		$token = $m[1];
		
		// Get the raw directory listing
		$directoryEscaped = urlencode($directory);
		$data = $this->dropbox_request('https://www.dropbox.com/browse2/'.$directoryEscaped.'?ajax=yes', true, array('t' => $token), true);

		// Post-condition the data so that it becomes valid XML
		$data = str_replace('&nbsp;', ' ', $data);

		$noScriptPos = strpos($data, '<noscript>');
		$data = substr($data, 0, $noScriptPos - 1);
		$data = '<'.'?xml version="1.0" encoding="UTF-8"?'.'><root>' . $data . '</root>';

		// Load the raw list as an XML document
		$xmlDoc = new DOMDocument();
		$xmlDoc->loadXML($data, LIBXML_NOBLANKS | LIBXML_NOCDATA | LIBXML_NOENT | LIBXML_NONET);

		$rootNodes = $xmlDoc->getElementsByTagname('root');
		$root = $rootNodes->item(0);

		$files = array();
		$folders = array();

		// Don't ask! This is my secret, deep magic approach to fetching the
		// DropBox directory listing.
		$childNodes = $root->childNodes;
		if($childNodes->length) for( $i = 0; $i < $childNodes->length; $i++ )
		{
			$thisNode = $childNodes->item($i);
			if($thisNode->nodeType != XML_ELEMENT_NODE) continue;
			if($thisNode->tagName !== 'div') continue;

			$name = $thisNode->attributes->getNamedItem('name')->textContent;
			switch($name) {
				case 'file row':
					$fileAnchor = $thisNode->childNodes->item(0)
						->childNodes->item(2)
						->childNodes->item(0);
					$sizeAnchor = $thisNode->childNodes->item(0)
						->childNodes->item(3);
					$modAnchor = $thisNode->childNodes->item(0)
						->childNodes->item(4);
					$files[$fileAnchor->textContent] = (object)array(
						'filename'	=> $fileAnchor->textContent,
						'url'		=> $fileAnchor->attributes->getNamedItem('href')->textContent,
						'size'		=> $sizeAnchor->textContent,
						'modified'	=> $modAnchor->textContent
					);
					break;

				case 'folder row':
					$fileAnchor = $thisNode->childNodes->item(0)
						->childNodes->item(2)
						->childNodes->item(0);
					$folders[$fileAnchor->textContent] = (object)array(
						'filename'	=> $fileAnchor->textContent,
						'url'		=> $fileAnchor->attributes->getNamedItem('href')->textContent
					);
					break;

				default:
					break;
			}
		}
		
		return (object)array('files'	=> $files, 'folders' => $folders);
	}
	
	/**
	 * Download using a signed download URL
	 * 
	 * @param type $url
	 * @param type $filename
	 * @param type $from
	 * @param type $length 
	 */
	public function dropbox_download($url, $filename, $from = null, $length = null) {
		$ch = curl_init();
		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		curl_setopt($ch, CURLOPT_HEADER, 0);
		@curl_setopt($ch, CURLOPT_FOLLOWLOCATION, 1);
		@curl_setopt($ch, CURLOPT_MAXREDIRS, 100);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		
		// Send cookies
		$rawCookies = array();
		foreach ($this->cookies as $k=>$v)
			$rawCookies[] = "$k=$v";
		$rawCookies = implode(';', $rawCookies);
		curl_setopt($ch, CURLOPT_COOKIE, $rawCookies);

		if(!is_null($from) && !is_null($length)) {
			curl_setopt($ch, CURLOPT_RANGE, $from.'-'.($from+$length-1));
		} elseif(!is_null($from)) {
			curl_setopt($ch, CURLOPT_RANGE, $from.'-');
		} elseif(!is_null($length)) {
			curl_setopt($ch, CURLOPT_RANGE, '0-'.($length-1));
		}

		$fp = fopen($filename,'wb');
		curl_setopt($ch, CURLOPT_FILE, $fp);

		$result = curl_exec($ch);
		curl_close($ch);
		fclose($fp);
	}
}